module Guitar
  class Fretboard
    include Common
    include TonicShape

    attr_reader :size, :strings, :changes
    attr_accessor :desc

    def initialize(size = DEFAULT_SIZE, desc: nil)
      @size = size
      @desc = desc
      @changes = []
      @strings = [
        String.new(:E4, Graphic.get_frets(1, size)),
        String.new(:B3, Graphic.get_frets(2, size)),
        String.new(:G3, Graphic.get_frets(3, size)),
        String.new(:D3, Graphic.get_frets(4, size)),
        String.new(:A2, Graphic.get_frets(5, size)),
        String.new(:E2, Graphic.get_frets(6, size))
      ]
    end

    def show_notes(notes = nil, string: nil, fret: nil, color: nil, &b)
      make_change do
        @strings.each_with_index do |s, i|
          if in?(i + 1, string)
            if block_given?
              s.show_notes(notes, fret: fret, color: color) do |n, f|
                b.call(n, i + 1, f)
              end
            else
              s.show_notes(notes, fret: fret, color: color)
            end
          end
        end
      end
    end

    def mark(content = nil, string: nil, fret: nil, color: nil)
      make_change do
        @strings.each_with_index do |s, i|
          s.mark(content, fret: fret, color: color) if in?(i + 1, string)
        end
      end
    end

    def play(note = nil, string: nil, fret: nil, color: nil)
      make_change do
        if string
          @strings[string - 1].play(note, fret: fret, color: color)
        else
          # TODO find string to play note
          fail 'no string or note to play'
        end
      end
    end

    def undo
      if last_change = changes.pop
        last_change.each { |i| @strings[i].undo }
      end

      self
    end

    def clear
      @strings.each(&:clear)
      changes.clear
      self
    end

    def inspect
      string_lines = @strings.collect(&:inspect).join("\n")
      baseline     = Graphic.get_baseline(@size)

      "#@desc\n#{string_lines}\n  #{baseline}"
    end

    private

    def record_change(&b)
      before = @strings.collect { |s| s.changes.size }
      b.call
    ensure
      after = @strings.collect { |s| s.changes.size }

      change = []
      after.each_with_index { |e, i| (e - before[i]).times { change << i } }
      changes << change if change.size > 0
    end
  end
end
