module Guitar
  class String
    include Common

    attr_reader :tonic, :notes, :frets, :changes

    def initialize(tonic, frets = Graphic.get_frets)
      @tonic   = tonic.to_n
      @frets   = frets
      @notes   = size.times.inject([@tonic]) { |r, i| r << (@tonic + i + 1) }
      @changes = []
    end

    def size
      frets.size
    end

    def show_notes(notes = nil, fret: nil, color: nil, &b)
      make_change do
        if notes
          show_given_notes(notes, fret, color, &b)
        else
          show_all_notes(fret, color, &b)
        end
      end
    end

    def mark(content = nil, fret: nil, color: nil)
      make_change do
        size.times { |i| mark_fret(i + 1, content, color) if in?(i + 1, fret) }
      end
    end

    PLAY_MARK = '◉ '

    def play(note = nil, fret: nil, color: nil)
      if fret
        Play.play(notes[fret])
        mark(PLAY_MARK, fret: fret, color: color)
      else
        note ||= tonic
        Play.play(note)

        fret   = notes.find_index { |n| n == note }
        fret ||= notes.find_index { |n| n === note }
        mark(PLAY_MARK, fret: fret, color: color)
      end
    end

    def undo
      if changes.size > 0
        @frets = changes.pop
      end

      self
    end

    def clear
      if changes.size > 0
        @frets = changes[0]
        changes.clear
      end

      self
    end

    def inspect
      "#{@tonic.name} #{@frets.join('')}"
    end

    private

    def mark_fret(i, e, color = nil)
      if i > 0
        i -= 1

        if e
          e = e.to_s
          s = e.uncolorize.size
          if s > Graphic::FRET_WIDTH
            fail "#{e} is too long, the size should < #{Graphic::FRET_WIDTH}"
          end

          start_index = (Graphic::FRET_WIDTH - s + 2) / 2
          end_index   = (-Graphic::FRET_WIDTH + s) / 2

          @frets[i] = (@changes[0] ? @changes[0][i] : @frets[i]).dup
          @frets[i][start_index..end_index] = colorize(e, color)
        else
          @frets[i] = colorize(@frets[i], color)
        end
      end
    end

    def record_change(&b)
      old = @frets.dup
      b.call
    ensure
      @changes << old if @frets != old
    end

    def show_all_notes(fret, color, &b)
      @notes.each_with_index do |e, i|
        if in?(i, fret)
          m = block_given? ? b.call(e, i) : e.name
          mark_fret(i, m, color)
        end
      end
    end

    def show_given_notes(notes, fret, color, &b)
      notes = notes.respond_to?(:map) ? notes.map(&:to_n) : [notes.to_n]

      @notes.each_with_index do |e, i|
        if in?(i, fret) && note = notes.find { |n| e === n }
          m = block_given? ? b.call(note, i) : note.name
          mark_fret(i, m, color)
        end
      end
    end
  end
end
