module Guitar
  module TonicShape
    # E ╓─────┬─────┬─────┬─────┬─────┬─────┬─────┬──Ⓒ ─┬─────┬─────┬─────┬────
    # B ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # G ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # D ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼──Ⓒ ─┼─────┼────
    # A ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # E ╙─────┴─────┴─────┴─────┴─────┴─────┴─────┴──Ⓒ ─┴─────┴─────┴─────┴────
    #                  𝇈           𝇈           𝇈           𝇈                𝇈 𝇈
    SHAPE_RIGHT_TRIANGLE = '>'

    # E ╓─────┬─────┬─────┬─────┬─────┬─────┬─────┬──Ⓒ ─┬─────┬─────┬─────┬────
    # B ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # G ╟─────┼─────┼─────┼─────┼──Ⓒ ─┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # D ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # A ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # E ╙─────┴─────┴─────┴─────┴─────┴─────┴─────┴──Ⓒ ─┴─────┴─────┴─────┴────
    #                  𝇈           𝇈           𝇈           𝇈                𝇈 𝇈
    SHAPE_LEFT_TRIANGLE = '<'

    # E ╓─────┬─────┬─────┬─────┬─────┬─────┬─────┬─────┬─────┬─────┬─────┬────
    # B ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # G ╟─────┼─────┼─────┼─────┼──Ⓒ ─┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # D ╟─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # A ╟─────┼─────┼──Ⓒ ─┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼─────┼────
    # E ╙─────┴─────┴─────┴─────┴─────┴─────┴─────┴─────┴─────┴─────┴─────┴────
    #                  𝇈           𝇈           𝇈           𝇈                𝇈 𝇈
    SHAPE_SLASH_BEFORE_TRIANGLE = '/<'

    # E ┄┄┬─────┬─────┬─────╥─────┬──
    # B ┄┄┼─────┼─────┼─────╫──Ⓒ ─┼──
    # G ┄┄┼─────┼─────┼─────╫─────┼──
    # D ┄┄┼──Ⓒ ─┼─────┼─────╫─────┼──
    # A ┄┄┼─────┼─────┼─────╫─────┼──
    # E ┄┄┴─────┴─────┴─────╨─────┴──
    #                   𝇈 𝇈
    SHAPE_SLASH_BEFORE_BACKSLASH = '/\\'

    # E ┄┄┄┄┄╥─────┬─────┬─────┬─────
    # B ┄┄┄┄┄╫──Ⓒ ─┼─────┼─────┼─────
    # G ┄┄┄┄┄╫─────┼─────┼─────┼─────
    # D ┄┄┄┄┄╫─────┼─────┼─────┼─────
    # A ┄┄┄┄┄╫─────┼─────┼──Ⓒ ─┼─────
    # E ┄┄┄┄┄╨─────┴─────┴─────┴─────
    #    𝇈 𝇈                𝇈
    SHAPE_BACKSLASH = '\\'

    AS_PRETTY_NAME = :pretty_name
    AS_DEGREE      = :degree
    AS_FINGER      = :finger

    DEGREES = %w(⑴  ⑵  ⑶  ⑷  ⑸  ⑹  ⑺)
    FINGERS = %w(①  ②  ③  ④)

    # TODO choose scale: major, minor ...
    def show_tonic_shape(tonic, shape = nil, as: nil, color: [[:red, :bold], :yellow])
      scale  = tonic.to_n.scale

      case shape.to_s
      when SHAPE_RIGHT_TRIANGLE
        show_tonic_shape_right_triangle(as, scale, color)
      when SHAPE_LEFT_TRIANGLE
        show_tonic_shape_left_triangle(as, scale, color)
      when SHAPE_BACKSLASH
        show_tonic_shape_backslash(as, scale, color)
      when SHAPE_SLASH_BEFORE_TRIANGLE
        show_tonic_shape_slash_before_triangle(as, scale, color)
      when SHAPE_SLASH_BEFORE_BACKSLASH
        show_tonic_shape_slash_before_backslash(as, scale, color)
      else
        show_all_tonic_shapes(as, scale, color)
      end
    end

    private

    def show_all_tonic_shapes(as, scale, color)
      show_notes(scale.notes) do |n|
        get_note_in_tonic_shape(n, as, scale, color)
      end
    end

    def show_tonic_shape_right_triangle(as, scale, color)
      fret = strings[5].notes.find_index { |n| n === scale.tonic }
      fret += 12 if fret < 2

      show_notes(scale.notes, fret: (fret - 1)..(fret + 2)) do |n, _, f|
        get_note_in_tonic_shape(n, as, scale, color, f - fret + 1)
      end
    end

    def show_tonic_shape_left_triangle(as, scale, color)
      fret = strings[5].notes.find_index { |n| n === scale.tonic }
      fret += 12 if fret < 5

      show_notes(scale.notes, fret: (fret - 4)..fret) do |n, s, f|
        get_note_in_tonic_shape(n, as, scale, color,
                                get_finger_in_left_triangle(s, fret, f))
      end
    end

    def get_finger_in_left_triangle(string, root_fret, current_fret)
      if string == 3
        current_fret - root_fret + 4
      else
        current_fret - root_fret + 3
      end
    end

    def show_tonic_shape_backslash(as, scale, color)
      fret = strings[4].notes.find_index { |n| n === scale.tonic }
      fret += 12 if fret < 4

      show_notes(scale.notes, fret: (fret - 3)..fret) do |n, _, f|
        get_note_in_tonic_shape(n, as, scale, color, f - fret + 3)
      end
    end

    def show_tonic_shape_slash_before_triangle(as, scale, color)
      fret = strings[4].notes.find_index { |n| n === scale.tonic }
      fret += 12 if fret < 2

      show_notes(scale.notes, fret: (fret - 1)..(fret + 3)) do |n, s, f|
        get_note_in_tonic_shape(n, as, scale, color,
                                get_finger_in_slash_before_triangle(s, fret, f))
      end
    end

    def get_finger_in_slash_before_triangle(string, root_fret, current_fret)
      if string < 3
        current_fret - root_fret
      else
        current_fret - root_fret + 1
      end
    end

    def show_tonic_shape_slash_before_backslash(as, scale, color)
      fret = strings[3].notes.find_index { |n| n === scale.tonic }
      fret += 12 if fret < 2

      show_notes(scale.notes, fret: (fret - 1)..(fret + 3)) do |n, s, f|
        get_note_in_tonic_shape(n, as, scale, color,
                                get_finger_in_slash_before_backslash(s, fret, f))
      end
    end

    def get_finger_in_slash_before_backslash(string, root_fret, current_fret)
      if string < 3 || string > 4
        current_fret - root_fret
      else
        current_fret - root_fret + 1
      end
    end

    def get_note_in_tonic_shape(note, as, scale, color, finger = nil)
      str = case as && as.to_sym
            when nil, AS_PRETTY_NAME
              note.pretty_name
            when AS_DEGREE
              DEGREES[scale.degree(note) - 1] + ' '
            when AS_FINGER
              finger ? FINGERS[finger] + ' ' : '⓿ '
            else
              note.name
            end

      color = [color] unless color.is_a?(Array)
      color = note === scale.tonic ? color[0] : color[1]

      colorize(str, color)
    end
  end
end
