module Guitar
  class Note
    include Comparable

    DEFAULT_OCTAVE = 4

    # http://newt.phys.unsw.edu.au/jw/notes.html
    MIDI_NUMBERS = {
      'C'  => 60,
      'C♯' => 61,
      'D♭' => 61,
      'D'  => 62,
      'D♯' => 63,
      'E♭' => 63,
      'E'  => 64,
      'F'  => 65,
      'F♯' => 66,
      'G♭' => 66,
      'G'  => 67,
      'G♯' => 68,
      'A♭' => 68,
      'A'  => 69,
      'A♯' => 70,
      'B♭' => 70,
      'B'  => 71
    }

    attr_reader :name, :octave, :number
    alias_method :to_i, :number

    NOTE_PATTERN  = /\A([a-gA-G][#s♯b♭]?)(\d)?\z/
    NATURAL_NOTES = %w(A B C D E F G).freeze

    # A0..C8
    def self.random(natural = false)
      if natural
        new("#{NATURAL_NOTES.sample}#{rand(0..8)}")
      else
        new(rand(21..108)).tap { |n| n.alias! if rand(0..1) == 0 }
      end
    end

    def initialize(note)
      case note
      when Integer
        init_with_number(note)
      else
        note = note.to_s

        if note =~ NOTE_PATTERN
          init_with_name_and_octave($1, $2 || DEFAULT_OCTAVE)
        else
          fail "invalid note #{note}"
        end
      end
    end

    def -(other)
      case other
      when Note
        @number - other.number
      when Integer
        self.class.new(@number - other)
      else
        fail "invalid operation with #{other}"
      end
    end

    def +(other)
      case
      when Integer
        self.- (-other)
      else
        fail "invalid operation with #{other}"
      end
    end

    def <=>(other)
      if other.respond_to?(:to_n)
        to_i <=> other.to_n.to_i
      else
        super
      end
    end

    def ===(other)
      if other.is_a?(Note)
        (self - other) % 12 == 0
      else
        super
      end
    end

    def inspect
      "#{@name}#{@octave}"
    end

    PRETTY_NAMES = {
      'C' => 'Ⓒ ',
      'D' => 'Ⓓ ',
      'E' => 'Ⓔ ',
      'F' => 'Ⓕ ',
      'G' => 'Ⓖ ',
      'A' => 'Ⓐ ',
      'B' => 'Ⓑ '
    }

    def pretty_name
      if @name.size == 1
        PRETTY_NAMES[@name]
      else
        "#{PRETTY_NAMES[@name[0]].strip}#{@name[1..-1]}"
      end
    end

    def alias!
      new_name, _ = MIDI_NUMBERS.find do |k, v|
        n = 60 + (@number % 12)
        v == n && k != name
      end

      @name = new_name if new_name
    end

    def scale(name = :major)
      Scale.new(self, name)
    end

    def to_n
      self
    end

    def play
      Play.play(self)
    end

    private

    def init_with_name_and_octave(name, octave)
      @name   = name.to_s.capitalize.sub(/[#s]/, '♯').sub('b', '♭')
      @octave = octave.to_i

      if number = MIDI_NUMBERS[@name]
        @number = number + (@octave - DEFAULT_OCTAVE) * 12
      else
        fail "the note name #{name} is not valid"
      end
    end

    def init_with_number(number)
      octave = ((number - 60) / 12) + DEFAULT_OCTAVE
      name   = MIDI_NUMBERS.key(60 + (number % 12))

      init_with_name_and_octave(name, octave)
    end
  end
end

[Integer, Symbol, String].each do |klass|
  klass.class_eval do
    def to_n
      Guitar::Note.new(self)
    end
  end
end
