module Guitar
  class Practice
    attr_reader :size, :interval, :natural, :color, :string, :fret

    def initialize(size: DEFAULT_SIZE, interval: 1, natural: false, color: nil,
                   string: 1..6, fret: 1..size)
      @size     = size
      @interval = interval
      @natural  = natural
      @color    = color
      @string   = string.to_a
      @fret     = fret.to_a
    end

    def random_note
      loop_and_print { Note.random(natural).name }
    end

    def random_note_in_random_string
      loop_and_print { "#{Note.random(natural).name} in string #{string.sample}" }
    end

    def random_note_in_random_fret
      loop_and_print { "#{Note.random(natural).name} in fret #{fret.sample}" }
    end

    def random_note_in_fretboard
      f = Guitar::Fretboard.new(size)
      prompt = 'Please input the current note: '

      loop do
        clear_screen
        f.clear

        note, string = Note.random(natural), self.string.sample
        puts f.show_notes(note, string: string, color: color) { '◉ ' }.inspect
        print prompt

        begin
          answer = STDIN.gets.strip

          if Note.new(answer) === note
            clear_screen
            puts f.show_notes(note, string: string, color: :green).inspect
            print prompt, answer
            sleep interval
          else
            print_error(f, note, string, prompt + answer)
          end
        rescue
          print_error(f, note, string, prompt + answer)
        end
      end
    end

    def mark_random_note_in_fretboard
      f = Guitar::Fretboard.new(size)

      loop do
        clear_screen
        f.clear

        note = Note.random(natural)
        prompt = "Please input the position of #{note.name}: "
        puts f.inspect
        print prompt

        # Show notes, if we get the right answer, they will be covered.
        f.show_notes(note, color: color)

        begin
          answer  = STDIN.gets.strip
          correct = answer != ''

          answer.split(/[\s,;]+/).each do |position|
            string, fret = position.split(':').map(&:to_i)
            f.show_notes(string: string, fret: fret) do |n|
              if n === note
                n.name.green
              else
                correct = false
                n.name.red
              end
            end
          end

          clear_screen
          puts f.inspect
          print prompt, answer

          if correct
            sleep interval
          else
            print "\nPress ENTER to continue ..."
            STDIN.gets
          end
        rescue
          clear_screen
          puts f.inspect
          print prompt, answer
          print "\nPress ENTER to continue ..."
          STDIN.gets
        end

      end
    end

    private

    def clear_screen
      print "\x1b[2J\x1b[1;1H"
    end

    def loop_and_print(&b)
      loop do
        clear_screen
        current = b.call
        current = current.send(color) if color && current.respond_to?(color)

        if current != @last
          print @last = current
          sleep interval
        end
      end
    end

    def print_error(fretboard, note, string, answer)
      clear_screen
      puts fretboard.show_notes(note, string: string, color: :red).inspect
      puts answer
      print 'Press ENTER to continue ...'
      STDIN.gets
    end
  end
end
